"""Lazy lemmings.

Find the farthest any single lemming needs to travel for food.

    >>> furthest(3, [0, 1, 2])
    0

    >>> furthest(3, [2])
    2

    >>> furthest(3, [0])
    2

    >>> furthest(6, [2, 4])
    2

    >>> furthest(7, [0, 6])
    3

    >>> furthest_optimized(7, [0, 6])
    3

    >>> furthest_optimized(3, [0, 1, 2])
    0

    >>> furthest_optimized(3, [2])
    2

    >>> furthest_optimized(3, [0])
    2

    >>> furthest_optimized(6, [2, 4])
    2
"""


def furthest(num_holes, cafes):
    """Find longest distance between a hole and a cafe."""

    # START SOLUTION

    worst = 0

    for hole in range(num_holes):
        # Looking at all cafes, find distance to this hole,
        # and choose the smallest distance.

        dist = min([abs(hole - cafe) for cafe in cafes])

        # Keep track of the longest distance we've seen

        worst = max(worst, dist)

    return worst


def furthest_optimized(num_holes, cafes):
    """Find longest distance between a hole and a cafe.

    This solution uses binary search to quickly find the one or two cafes
    closest to the hole. This makes problem O(n log n), rather than O(n**2).
    """

    from bisect import bisect_left

    worst = 0

    for hole in range(num_holes):

        # Find the place we'd insert this hole into the
        # sorted cafes list.

        idx = bisect_left(cafes, hole)

        if idx == len(cafes):
            # This hole is after all the cafes, so the distance
            # is from this hole to the cafe before it
            dist = hole - cafes[idx - 1]

        elif idx == 0:
            # This hole is before all the cafes, so the distance
            # is from this hole to the cafe after it
            dist = cafes[idx] - hole

        elif cafes[idx] == hole:
            # This hole is a cafe, so no travel is needed!
            dist = 0

        else:
            # This hole is between two cafes, so use the smaller
            # distance between them
            dist = min(hole - cafes[idx - 1], cafes[idx] - hole)

        # Keep track of the longest distance we've seen
        worst = max(worst, dist)

    return worst

# If you'd like to bypass Python's bisect_left in order to
# understand the Binary Search aspect of the above solution,
# check out this version.
# You could use the below function instead of bisect.bisect_left.
def another_bisect_left(ordered_iterable, thing_to_insert):
    """Mimes the behavior of Python's bisect_left function.

    This is a re-written, slightly simpified version of the source code
    for bisect.bisect_left.
    """

    lower_bound = 0
    upper_bound = len(ordered_iterable)
    while lower_bound < upper_bound:
        mid = (lower_bound+upper_bound) / 2
        if ordered_iterable[mid] < thing_to_insert: 
            lower_bound = mid+1
        else: 
            upper_bound = mid
    return lower_bound


def furthest_best(num_holes, cafes):
    """Find longest distance between a hole and a cafe.

    This solution looks at the cafes, finding the two that are furthest
    apart, and then determining which lemming hole would have to travel
    the furthest between them.
    """

    # Start with the distance from first hole to the first cafe, and from
    # the last hole to the last cafe:

    distances = [cafes[0],
                 num_holes - cafes[-1] - 1]

    for i in range(1, len(cafes)):
        # between cafes: distance is half the distance to leftward cafe
        distances.append((cafes[i] - cafes[i - 1]) // 2)

    return max(distances)


# END SOLUTION

if __name__ == '__main__':
    import doctest
    if doctest.testmod().failed == 0:
        print("\n*** ALL TESTS PASSED; GREAT JOB!\n")
